unit Dbgrid;
(*
 Enhanced DBGrid V1.06 - released September 18, 1997

 Disclaimer:
 This is a Freeware component. Use it at your own risk.
 You can send any comments, suggestions and bugs to demian@bhnet.com.br
 New versions can be found at: http://www.bhnet.com.br/~demian
                               http://web.unix.horizontes.com.br/~demian

 If you do any modification to this code, please send me a copy.
 If you use this code on any 'real-life' application, please let me know.
 If you think this code is completely useless, please DON'T LET ME KNOW!

 What it does:
 DBGrid descendent with support for embedded objects. Also publishs Col/Row properties
 and OnMouseDown/OnMouseMove/OnMouseUp events.

 How to use it (a sample for embedding a TComboBox)):
 1. Install the component [VNM Pallete];
 2. Drop one TEDBGrid and one TTable onto a form;
 3. Drop one TComboBox onto the TEDBGrid component
 4. Double click the TTable, define a new calculated string field, name it 'ComboBox1';

 Notice:
 - To disable/enable object drawing set the public property ObjEnabled to true/false;
 - Each embedded object must have a corresponding Calculated Field;
 - As default the object size and position will be the same of the current grid
   cell (ideal for TCustomEdit descendents). If you don't want the object size
   to be adjusted (if the object is a TDBMemo, for example), set the object Tag
   property to a value other than 0;
 - See 'My Defaults' comment in the Create method.

 History:

 1.06  - fixed OnMouseDown bug;
 1.05  - fixed persistent columns bug (Delphi 2 only);
 1.04  - Eliminated the ColObjects property and implemented objects dropping
         onto the grid component;
       - Altered objects displaying (now they are showed only when the user
         press a key or click the mouse over the embedded object grid cell);
       - fixed navigational problems when exiting an embedded object.
 1.03  - Fixed calculated fields navigational problems;
 1.02  - Fixed TDBCustomGrid memory leak (Delphi 1 only);
 1.01  - Published events OnMouseDown and OnMouseUp for dragging suport;
       - Fixed typo 'Miscelan' in Uses clause;
 1.00  - Created;
*)

interface

uses
  SysUtils, Classes, Grids, Controls, DB, DBGrids, Graphics, Forms, WinTypes, Messages;

type
  TEDBGrid = class(TDBGrid)
  private
    iCurrentObj: integer;
    ColumnHasObj: boolean;
    ShowObject: boolean;
    GridObjs: TList;
    FOnMouseDown: TMouseEvent;
    function GetObjIndex (const GridIndex: integer): integer;
    procedure WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
  protected
    procedure Loaded; override;
    procedure SetColumnAttributes; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer); override;
    procedure KeyPress(var Key: Char); override;
    {$ifdef WIN32}
    procedure DrawColumnCell(const Rect: TRect; DataCol: Integer;
      Column: TColumn; State: TGridDrawState); override;
    {$else}
    procedure DrawDataCell(const Rect: TRect; Field: TField; State: TGridDrawState); override;
    {$endif}
  public
    ObjEnabled: boolean;
    Property Row;
    Property Col;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
  published
    property OnMouseDown read FOnMouseDown write FOnMouseDown;
    Property OnMouseMove;
    Property OnMouseUp;
  end;

procedure Register;

implementation

{______________________________________________________________________________}
procedure Register;
begin
  RegisterComponents('VNM', [TEDBGrid]);
end;

{______________________________________________________________________________}
constructor TEDBGrid.Create (AOwner:TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csAcceptsControls];
  ObjEnabled := true;
  iCurrentObj := -1;                           {current visible embedded object}
  ShowObject := false;                    {show only when KeyPress or MouseDown}
  ColumnHasObj := false;
  GridObjs := TList.Create;
  {My Defaults (you may change or delete this to suit your needs)}
  Options := Options - [dgConfirmDelete,dgColumnResize];
  TitleFont.Style := [fsBold];
  Height := 142;
end;

{______________________________________________________________________________}
destructor TEDBGrid.Destroy;
begin
  inherited Destroy;
  GridObjs.Free;
  GridObjs := nil;
  {$IFNDEF WIN32}
  try
    TitleFont.Free;                                   {fix Delphi 1 memory leak}
    TitleFont:=Nil;
  except
  end;
  {$ENDIF}
end;

{______________________________________________________________________________}
procedure TEDBGrid.Loaded;
begin
  Inherited Loaded;
  if csDesigning in ComponentState then Exit;
  {move all objects from Controls property to GridObjs}
  {Notice: this is necessary because some controls (TComboBox, for example)
   don't "behave" themselves when having the grid as parent (Who knows why?)}
  while ControlCount > 0 do begin
    {hide all embedded objects}
    Controls[ControlCount-1].Visible := false;
    {move objects to GridObjs}
    GridObjs.Add(Controls[ControlCount-1]);
    TControl(GridObjs[GridObjs.Count-1]).Parent := Parent;
  end;
end;

{______________________________________________________________________________}
procedure TEDBGrid.SetColumnAttributes;
var
  i: Integer;
begin
  inherited SetColumnAttributes;
  {tell the grid to stop at each embedded object}
  for i := 1 to FieldCount do
    if not TabStops[i] then
      TabStops[i] := (GetObjIndex(i-1) <> -1);
end;

{______________________________________________________________________________}
function TEDBGrid.GetObjIndex (const GridIndex: integer): integer;
var
  i: integer;
begin
  result := -1;
  for i := 0 to GridObjs.Count-1 do
    {each embedded object must have a corresponding Calculated Field}
    if CompareText(TControl(GridObjs[i]).Name,Fields[GridIndex].FieldName) = 0 then begin
      result := i;
      break
    end;
end;

{______________________________________________________________________________}
procedure TEDBGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  if Assigned(FOnMouseDown) then
     FOnMouseDown(Self,Button,Shift,X,Y);
  inherited MouseDown(Button, Shift, X, Y);
  {show current embedded object}
  ShowObject := true;
  DrawCell(Col,Row,CellRect(Col,Row),[gdFocused,gdSelected]);
end;

{______________________________________________________________________________}
procedure TEDBGrid.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  {show current embedded object, if need be}
  if ColumnHasObj and (Key <> #9) then begin
    ShowObject := true;
    DrawCell(Col,Row,CellRect(Col,Row),[gdFocused]);
  end;
end;

{______________________________________________________________________________}
{$ifdef WIN32}
procedure TEDBGrid.DrawColumnCell(const Rect: TRect; DataCol: Integer;
                                  Column: TColumn; State: TGridDrawState);
{$else}
procedure TEDBGrid.DrawDataCell(const Rect: TRect; Field: TField;
                                State: TGridDrawState);
{$endif}
begin
  {$ifdef WIN32}
  inherited DrawColumnCell(Rect,DataCol,Column,State);
  {$else}
  inherited DrawDataCell(Rect,Field,State);
  {$endif}
  if (gdFocused in State) and ObjEnabled then begin
    {check if there's a visible object then hide it}
    if iCurrentObj <> -1 then begin
      TControl(GridObjs[iCurrentObj]).Visible := false;
      iCurrentObj := -1;
    end;
    {check if the current column has an enbedded object then show it}
    if ShowObject then begin
      iCurrentObj := GetObjIndex(SelectedIndex);
      if iCurrentObj <> -1 then with TControl(GridObjs[iCurrentObj]) do begin
        if Tag = 0 then
          Setbounds(Rect.Left+Self.Left,Rect.Top+Self.Top,Rect.Right-Rect.Left,Height)
        else
          Setbounds(Rect.Left+Self.Left,Rect.Top+Self.Top,Width,Height);
        {make it visible and give it focus}
        Visible := true;
        TWinControl(GridObjs[iCurrentObj]).SetFocus;
      end;
      ShowObject := false;
    end;
  end;
end;

{______________________________________________________________________________}
procedure TEDBGrid.WMSetFocus(var Msg: TWMSetFocus);
begin
  inherited;
  {adjust the cursor position after exiting an embedded object}
  if ColumnHasObj then
    if Col < FieldCount then
      Col := Col + 1
    else begin
      Col := 1;
      Datasource.Dataset.Next;
    end;
end;

{______________________________________________________________________________}
procedure TEDBGrid.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  inherited;
  {set flag for KeyPress event (called once when entering each grid column)}
  ColumnHasObj := GetObjIndex(Col-1) <> -1;
end;

end.

